@@ -258,7 +258,7 @@ do_export (BuilderContext *build_context,
   g_ptr_array_add (args, NULL);
 
   return flatpak_spawnv (NULL, NULL, G_SUBPROCESS_FLAGS_NONE, error,
-                         (const gchar * const *) args->pdata);
+                         (const gchar * const *) args->pdata, NULL);
 }
 
 static gboolean
@@ -297,7 +297,7 @@ do_install (BuilderContext *build_context,
   g_ptr_array_add (args, NULL);
 
   return flatpak_spawnv (NULL, NULL, G_SUBPROCESS_FLAGS_NONE, error,
-                         (const gchar * const *) args->pdata);
+                         (const gchar * const *) args->pdata, NULL);
 }
 
 static gboolean
@@ -981,7 +981,8 @@ main (int    argc,
                                           NULL,
                                           0,
                                           &error,
-                                          argv))
+                                          argv,
+                                          NULL))
             {
               g_printerr ("Error mirroring screenshots: %s\n", error->message);
               return 1;
@@ -1097,9 +1098,13 @@ main (int    argc,
                                        builder_context_get_build_runtime (build_context) ? "usr" : "files",
                                        builder_extension_get_directory (e));
 
+          const char *extension_branch = builder_extension_get_version (e);
+          if (extension_branch == NULL)
+            extension_branch = builder_manifest_get_branch (manifest, build_context);
+
           if (!do_export (build_context, &error, TRUE,
                           flatpak_file_get_path_cached (export_repo),
-                          app_dir_path, NULL, builder_manifest_get_branch (manifest, build_context),
+                          app_dir_path, NULL, extension_branch,
                           builder_manifest_get_collection_id (manifest),
                           builder_manifest_get_token_type (manifest),
                           metadata_arg, files_arg,
