@@ -46,6 +46,8 @@ struct BuilderModule
   char           *subdir;
   char          **post_install;
   char          **config_opts;
+  char          **secret_opts;
+  char          **secret_env;
   char          **make_args;
   char          **make_install_args;
   char           *install_rule;
@@ -99,6 +101,8 @@ enum {
   PROP_BUILDSYSTEM,
   PROP_BUILDDIR,
   PROP_CONFIG_OPTS,
+  PROP_SECRET_OPTS,
+  PROP_SECRET_ENV,
   PROP_MAKE_ARGS,
   PROP_MAKE_INSTALL_ARGS,
   PROP_ENSURE_WRITABLE,
@@ -144,6 +148,8 @@ builder_module_finalize (GObject *object)
   g_free (self->buildsystem);
   g_strfreev (self->post_install);
   g_strfreev (self->config_opts);
+  g_strfreev (self->secret_opts);
+  g_strfreev (self->secret_env);
   g_strfreev (self->make_args);
   g_strfreev (self->make_install_args);
   g_strfreev (self->ensure_writable);
@@ -229,6 +235,14 @@ builder_module_get_property (GObject    *object,
       g_value_set_boxed (value, self->config_opts);
       break;
 
+    case PROP_SECRET_OPTS:
+      g_value_set_boxed (value, self->secret_opts);
+      break;
+
+    case PROP_SECRET_ENV:
+      g_value_set_boxed (value, self->secret_env);
+      break;
+
     case PROP_MAKE_ARGS:
       g_value_set_boxed (value, self->make_args);
       break;
@@ -370,6 +384,18 @@ builder_module_set_property (GObject      *object,
       g_strfreev (tmp);
       break;
 
+    case PROP_SECRET_OPTS:
+      tmp = self->secret_opts;
+      self->secret_opts = g_strdupv (g_value_get_boxed (value));
+      g_strfreev (tmp);
+      break;
+
+    case PROP_SECRET_ENV:
+      tmp = self->secret_env;
+      self->secret_env = g_strdupv (g_value_get_boxed (value));
+      g_strfreev (tmp);
+      break;
+
     case PROP_MAKE_ARGS:
       tmp = self->make_args;
       self->make_args = g_strdupv (g_value_get_boxed (value));
@@ -568,6 +594,23 @@ builder_module_class_init (BuilderModuleClass *klass)
                                                        "",
                                                        G_TYPE_STRV,
                                                        G_PARAM_READWRITE));
+
+  g_object_class_install_property (object_class,
+                                   PROP_SECRET_OPTS,
+                                   g_param_spec_boxed ("secret-opts",
+                                                       "",
+                                                       "",
+                                                       G_TYPE_STRV,
+                                                       G_PARAM_READWRITE));
+
+  g_object_class_install_property (object_class,
+                                   PROP_SECRET_ENV,
+                                   g_param_spec_boxed ("secret-env",
+                                                       "",
+                                                       "",
+                                                       G_TYPE_STRV,
+                                                       G_PARAM_READWRITE));
+
   g_object_class_install_property (object_class,
                                    PROP_MAKE_ARGS,
                                    g_param_spec_boxed ("make-args",
@@ -787,6 +830,7 @@ load_sources_from_json (const char *sources_relpath)
   else
     return NULL;
 
+  builder_manifest_set_demarshal_base_dir (saved_demarshal_base_dir);
   return g_steal_pointer (&sources);
 }
 
@@ -1250,6 +1294,8 @@ shell (GFile          *app_dir,
 
 static const char skip_arg[] = "skip";
 static const char strv_arg[] = "strv";
+static const char secret_arg[] = "secret";
+static const char secret_env_arg[] = "secret_env";
 
 static gboolean
 build (GFile          *app_dir,
@@ -1263,16 +1309,49 @@ build (GFile          *app_dir,
        const gchar    *argv1,
        ...)
 {
+  g_auto(GStrv) env = g_strdupv (env_vars);
+  g_auto(GStrv) unresolved_env = g_strdupv (env_vars);
   g_autoptr(GFile) cwd_file = NULL;
-  g_autoptr(GPtrArray) args =
-    setup_build_args (app_dir, module_name, context, source_dir, cwd_subdir, flatpak_opts, env_vars, &cwd_file);
+  g_autoptr(GPtrArray) args = NULL;
+  g_autoptr(GPtrArray) unresolved_args = NULL;
+  gboolean have_secrets = FALSE;
+  gboolean build_success = FALSE;
   const gchar *arg;
   const gchar **argv;
   va_list ap;
   int i;
 
+  va_start (ap, argv1);
+  while ((arg = va_arg (ap, const gchar *)))
+    {
+      if (arg == secret_env_arg)
+        {
+          have_secrets = TRUE;
+          argv = va_arg (ap, const gchar **);
+          if (argv != NULL)
+            {
+              for (i = 0; argv[i] != NULL; i++)
+                {
+                  const char *secret_value = getenv (argv[i]);
+                  if (secret_value)
+                    {
+                      env = g_environ_setenv (g_strdupv (env), g_strdup (argv[i]), g_strdup (secret_value), FALSE);
+                      unresolved_env = g_environ_setenv (g_strdupv (unresolved_env), g_strdup (argv[i]), g_strdup_printf ("$host:%s", argv[i]), FALSE);
+                    }
+                }
+            }
+        }
+    }
+  va_end (ap);
+
+  args =
+    setup_build_args (app_dir, module_name, context, source_dir, cwd_subdir, flatpak_opts, env, &cwd_file);
+  unresolved_args =
+    setup_build_args (app_dir, module_name, context, source_dir, cwd_subdir, flatpak_opts, unresolved_env, &cwd_file);
+
   va_start (ap, argv1);
   g_ptr_array_add (args, g_strdup (argv1));
+  g_ptr_array_add (unresolved_args, g_strdup (argv1));
   while ((arg = va_arg (ap, const gchar *)))
     {
       if (arg == strv_arg)
@@ -1281,19 +1360,54 @@ build (GFile          *app_dir,
           if (argv != NULL)
             {
               for (i = 0; argv[i] != NULL; i++)
-                g_ptr_array_add (args, g_strdup (argv[i]));
+                {
+                  g_ptr_array_add (args, g_strdup (argv[i]));
+                  g_ptr_array_add (unresolved_args, g_strdup (argv[i]));
+                }
             }
         }
+      else if (arg == secret_arg)
+        {
+          argv = va_arg (ap, const gchar **);
+          if (argv != NULL)
+            {
+              have_secrets = TRUE;
+              for (i = 0; argv[i] != NULL; i++)
+                {
+                  const char *cur = strchr (argv[i], '$');
+                  if (cur)
+                    {
+                      g_autofree char *secret_key = g_strndup (argv[i], cur - argv[i]);
+                      const char *secret_value = getenv (cur + 1);
+                      if (secret_value)
+                        {
+                          g_autofree char *secret_opt = g_strconcat (secret_key, secret_value, NULL);
+                          g_ptr_array_add (args, g_strdup (secret_opt));
+                          g_ptr_array_add (unresolved_args, g_strdup_printf ("%s$host:%s", secret_key, cur + 1));
+                        }
+                    }
+                }
+            }
+        }
+      else if (arg == secret_env_arg)
+        va_arg (ap, const gchar **);
       else if (arg != skip_arg)
         {
           g_ptr_array_add (args, g_strdup (arg));
+          g_ptr_array_add (unresolved_args, g_strdup (arg));
         }
     }
   va_end (ap);
 
   g_ptr_array_add (args, NULL);
+  g_ptr_array_add (unresolved_args, NULL);
 
-  if (!builder_maybe_host_spawnv (cwd_file, NULL, 0, error, (const char * const *)args->pdata))
+  if (have_secrets)
+    build_success = builder_maybe_host_spawnv (cwd_file, NULL, 0, error, (const char * const *)args->pdata, (const char * const *)unresolved_args->pdata);
+  else
+    build_success = builder_maybe_host_spawnv (cwd_file, NULL, 0, error, (const char * const *)args->pdata, NULL);
+
+  if (!build_success)
     {
       g_prefix_error (error, "module %s: ", module_name);
       return FALSE;
@@ -1407,6 +1521,8 @@ builder_module_build_helper (BuilderModule  *self,
   g_auto(GStrv) env = NULL;
   g_auto(GStrv) build_args = NULL;
   g_auto(GStrv) config_opts = NULL;
+  g_auto(GStrv) secret_opts = NULL;
+  g_auto(GStrv) secret_env = NULL;
   g_autoptr(GFile) source_subdir = NULL;
   const char *source_subdir_relative = NULL;
   g_autofree char *source_dir_path = NULL;
@@ -1439,6 +1555,8 @@ builder_module_build_helper (BuilderModule  *self,
 
   env = builder_options_get_env (self->build_options, context);
   config_opts = builder_options_get_config_opts (self->build_options, context, self->config_opts);
+  secret_opts = builder_options_get_secret_opts (self->build_options, context, self->secret_opts);
+  secret_env = builder_options_get_secret_env (self->build_options, context, self->secret_env);
 
   n_jobs = g_strdup_printf ("%d", self->no_parallel_make ? 1 : builder_context_get_jobs (context));
   env = g_environ_setenv (env, "FLATPAK_BUILDER_N_JOBS", n_jobs, FALSE);
@@ -1678,7 +1796,7 @@ builder_module_build_helper (BuilderModule  *self,
       configure_args = (char **) g_ptr_array_free (g_steal_pointer (&configure_args_arr), FALSE);
 
       if (!build (app_dir, self->name, context, source_dir, build_dir_relative, build_args, env, error,
-                  configure_cmd, strv_arg, configure_args, strv_arg, config_opts, NULL))
+                  configure_cmd, strv_arg, configure_args, strv_arg, config_opts, secret_arg, secret_opts, NULL))
         return FALSE;
     }
   else
@@ -1765,7 +1883,7 @@ builder_module_build_helper (BuilderModule  *self,
     {
       g_print ("Running: %s\n", self->build_commands[i]);
       if (!build (app_dir, self->name, context, source_dir, build_dir_relative, build_args, env, error,
-                  "/bin/sh", "-c", self->build_commands[i], NULL))
+                  "/bin/sh", "-c", self->build_commands[i], secret_env_arg, secret_env, NULL))
         return FALSE;
     }
 
@@ -1803,7 +1921,7 @@ builder_module_build_helper (BuilderModule  *self,
       for (i = 0; self->post_install[i] != NULL; i++)
         {
           if (!build (app_dir, self->name, context, source_dir, build_dir_relative, build_args, env, error,
-                      "/bin/sh", "-c", self->post_install[i], NULL))
+                      "/bin/sh", "-c", self->post_install[i], secret_env_arg, secret_env, NULL))
             return FALSE;
         }
     }
@@ -1975,6 +2093,7 @@ builder_module_checksum (BuilderModule  *self,
   builder_cache_checksum_str (cache, self->subdir);
   builder_cache_checksum_strv (cache, self->post_install);
   builder_cache_checksum_strv (cache, self->config_opts);
+  builder_cache_checksum_strv (cache, self->secret_opts);
   builder_cache_checksum_strv (cache, self->make_args);
   builder_cache_checksum_strv (cache, self->make_install_args);
   builder_cache_checksum_strv (cache, self->ensure_writable);
