@@ -272,6 +272,19 @@ builder_source_archive_set_property (GObject      *object,
     }
 }
 
+static gboolean
+builder_source_archive_validate (BuilderSource  *source,
+                                 GError        **error)
+{
+  BuilderSourceArchive *self = BUILDER_SOURCE_ARCHIVE (source);
+
+  if (self->dest_filename != NULL &&
+      strchr (self->dest_filename, '/') != NULL)
+    return flatpak_fail (error, "No slashes allowed in dest-filename, use dest property for directory");
+
+  return TRUE;
+}
+
 static SoupURI *
 get_uri (BuilderSourceArchive *self,
          GError              **error)
@@ -472,7 +485,7 @@ un7z (GFile       *dir,
   gboolean res;
   const gchar *argv[] = { "7z",  "x", sevenz_path, NULL };
 
-  res = flatpak_spawnv (dir, NULL, 0, error, argv);
+  res = flatpak_spawnv (dir, NULL, 0, error, argv, NULL);
 
   return res;
 }
@@ -488,7 +501,7 @@ unrpm (GFile   *dir,
       rpm_path, /* shell's $1 */
       NULL };
 
-  res = flatpak_spawnv (dir, NULL, 0, error, argv);
+  res = flatpak_spawnv (dir, NULL, 0, error, argv, NULL);
 
   return res;
 }
@@ -692,6 +705,7 @@ get_type_from_prop (BuilderSourceArchive *self)
 static gboolean
 builder_source_archive_extract (BuilderSource  *source,
                                 GFile          *dest,
+                                GFile          *source_dir,
                                 BuilderOptions *build_options,
                                 BuilderContext *context,
                                 GError        **error)
@@ -873,6 +887,7 @@ builder_source_archive_class_init (BuilderSourceArchiveClass *klass)
   source_class->extract = builder_source_archive_extract;
   source_class->bundle = builder_source_archive_bundle;
   source_class->checksum = builder_source_archive_checksum;
+  source_class->validate = builder_source_archive_validate;
 
   g_object_class_install_property (object_class,
                                    PROP_PATH,
